#pragma once

// Genode includes
#include <base/heap.h>
#include <libc/component.h>
#include <util/list.h>

// gapfruit inculdes
#include <util/hexdump.h>
#include <util/vector.h>

// local includes
#include "command.h"
#include "error.h"
#include "exception.h"
#include "payload.h"


namespace Hsm {
	enum {
		PROTOCOL_VERSION = 0x00050000,
	};

	struct Header;
	class  Message;
}

struct __attribute__ ((packed)) Hsm::Header
{
	uint32_t protocol_version;
	uint32_t data_length;
	uint32_t command_id;
	uint32_t session_id;
	uint32_t network_transaction_number;
	uint32_t hsm_data_generation_number;
	uint32_t status;
	uint32_t reserved0;
	uint32_t reserved1;
	uint32_t reserved2;
};
class Hsm::Message
{
	private:
		Gapfruit::Vector<uint8_t> _data;

		Header const &_header() const { return *reinterpret_cast<Header const *>(_data.data()); }
		Header &_header() { return *reinterpret_cast<Header *>(_data.data()); }

		uint8_t const *payload() const { return &_data[sizeof(Header)]; }
		uint8_t *payload() { return &_data[sizeof(Header)]; }

		size_t payload_size() const {
			return _header().data_length;
		}

	public:
		Message(Message&& other) = default;

		Message(Genode::Heap& heap, const Command command)
			: _data(sizeof(Header), 0, heap)
		{
			_header().protocol_version = PROTOCOL_VERSION;
			_header().command_id = command;
		}

		template <typename SOCKET>
		Message(Genode::Heap& heap, const SOCKET& socket)
			: _data(sizeof(Header), heap)
		{
			size_t result = socket.read(_data.data(), sizeof(Header));
			if (result < sizeof(Header)) {
				Genode::error("Failed to read HSM header: got ", result,
				              " bytes, expecting ", sizeof(Header));
				throw Error_response_exception { };
			}
			if (_header().protocol_version != PROTOCOL_VERSION) {
				// if version do not match it also maybe because HSM has sent to us
				// all features it suppports in the first byte, because max connections
				// number exceeded
				constexpr uint32_t const mask = ~0xFFu;
				if ((_header().protocol_version & mask) == (PROTOCOL_VERSION & mask) &&
				    _header().status == ERR_MAX_CONNECTIONS_EXCEEDED) {
					// we must handle this situation at the connection level
					return;
				}
				Genode::error("Bad HSM protocol version. Expected ",
				              static_cast<uint32_t>(PROTOCOL_VERSION),
				              ", actual ",
				              static_cast<uint32_t>(_header().protocol_version));
				throw Bad_hsm_protcol_version_exception();
			}

			if (_header().data_length > 0) {
				_data.resize(sizeof(Header) + _header().data_length);
				socket.read(payload(), _header().data_length);
			}
		}

		Command command() const { return static_cast<Command>(_header().command_id); }
		uint32_t status() const { return _header().status; }

		Payload payload_by_type(const Data_type data_type) const
		{
			auto current = Payload { payload() };
			for (; current.raw() < _data.data() + _data.size(); current = current.next()) {
				if (current.header().type == data_type)
					return current;
			}
			return Payload { };
		}

		uint32_t data_length() const
		{
			return _header().data_length;
		}

		uint32_t size() const
		{
			return _data.size();
		}

		void add(const Data_type data_type, const uint8_t* payload_buffer, const uint32_t payload_length)
		{
			_data.resize(_data.size() + Payload_helper::padded_size(payload_length));
			Mutable_payload p { payload() + _header().data_length };
			_header().data_length += p.update(data_type, payload_buffer, payload_length);
		}

		void add(const Data_type data_type, const char* value)
		{
			add(data_type, reinterpret_cast<const uint8_t*>(value), strlen(value));
		}

		template <typename VALUE>
		void add(const Data_type data_type, const VALUE &value)
		{
			// FIXME: don't we need to consider endianess here?!
			add(data_type, reinterpret_cast<const uint8_t*>(&value), sizeof(VALUE));
		}

		uint8_t const* data() const { return _data.data(); }
		uint8_t* data() { return _data.data(); }
};

